(* Copyright 2011 Cornell University
 * Copyright 2012 Cornell University
 *
 *
 * This file is part of EventML - a tool aiming at specifying
 * distributed protocols in an ML like language.  It is an interface
 * to the logic of events and is compiled into Nuprl.  It is written
 * by the NUPRL group of Cornell University, Ithaca, NY.
 *
 * EventML is a free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * EventML is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with EventML.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  o Authors:     Vincent Rahli
 *  o Affiliation: Cornell University, NUPRL group
 *  o Date:        20 May 2011
 *  o File name:   Interface.sml
 *  o Description: EventML interface.
 *)


structure Interface :> INTERFACE = struct

structure P  = Parser
structure C  = ConfigParse
structure A  = Ast
structure E  = Enum
structure N  = ToNuprl
structure D  = Deps
structure G  = Gen
structure EN = Env
structure GL = GenLib
structure NT = NuprlTerms
structure EH = LibBase
structure EV = Evaluators
structure PN = ParserNuprlAscii


(* ------ ARGUMENTS ------ *)

val default_el_output = "/tmp/eventml-output.el"
val default_output    = NONE
val default_lib       = NONE
val default_input     = "test.esh"
val default_time      = LargeInt.fromInt 1000 (* 1sec by default *)
val default_sub       = false
val default_sanity    = false
val default_nuprl     = false
val default_obid      = ""
val default_ascii     = false
val default_tcheck    = false
val default_parse     = false
val default_split     = false
val default_prt       = false
val default_eval      = NONE
val default_alldef    = NONE
val default_test      = NONE
val default_session   = false
val default_simul     = NONE
val default_step      = NONE
val default_mono      = false

val initArgs =
    {input   = default_input,
     output  = default_output,
     lib     = default_lib,
     time    = default_time,
     sub     = default_sub,
     sanity  = default_sanity,
     nuprl   = default_nuprl,
     obid    = default_obid,
     ascii   = default_ascii,
     tcheck  = default_tcheck,
     parse   = default_parse,
     split   = default_split,
     prt     = default_prt,
     eval    = default_eval,
     alldef  = default_alldef,
     test    = default_test,
     session = default_session,
     simul   = default_simul,
     step    = default_step,
     mono    = default_mono}

fun getElOutput (SOME output) = output
  | getElOutput NONE = default_el_output

fun updInput   {input = _, output, lib, time, sub, sanity, nuprl, obid, ascii, tcheck, parse, split, prt, eval, alldef, test, session, simul, step, mono} input   = {input = input, output = output, lib = lib, time = time, sub = sub, sanity = sanity, nuprl = nuprl, obid = obid, ascii = ascii, tcheck = tcheck, parse = parse, split = split, prt = prt, eval = eval, alldef = alldef, test = test, session = session, simul = simul, step = step, mono = mono}
fun updOutput  {input, output = _, lib, time, sub, sanity, nuprl, obid, ascii, tcheck, parse, split, prt, eval, alldef, test, session, simul, step, mono} output  = {input = input, output = output, lib = lib, time = time, sub = sub, sanity = sanity, nuprl = nuprl, obid = obid, ascii = ascii, tcheck = tcheck, parse = parse, split = split, prt = prt, eval = eval, alldef = alldef, test = test, session = session, simul = simul, step = step, mono = mono}
fun updLib     {input, output, lib = _, time, sub, sanity, nuprl, obid, ascii, tcheck, parse, split, prt, eval, alldef, test, session, simul, step, mono} lib     = {input = input, output = output, lib = lib, time = time, sub = sub, sanity = sanity, nuprl = nuprl, obid = obid, ascii = ascii, tcheck = tcheck, parse = parse, split = split, prt = prt, eval = eval, alldef = alldef, test = test, session = session, simul = simul, step = step, mono = mono}
fun updTime    {input, output, lib, time = _, sub, sanity, nuprl, obid, ascii, tcheck, parse, split, prt, eval, alldef, test, session, simul, step, mono} time    = {input = input, output = output, lib = lib, time = time, sub = sub, sanity = sanity, nuprl = nuprl, obid = obid, ascii = ascii, tcheck = tcheck, parse = parse, split = split, prt = prt, eval = eval, alldef = alldef, test = test, session = session, simul = simul, step = step, mono = mono}
fun updSub     {input, output, lib, time, sub = _, sanity, nuprl, obid, ascii, tcheck, parse, split, prt, eval, alldef, test, session, simul, step, mono} sub     = {input = input, output = output, lib = lib, time = time, sub = sub, sanity = sanity, nuprl = nuprl, obid = obid, ascii = ascii, tcheck = tcheck, parse = parse, split = split, prt = prt, eval = eval, alldef = alldef, test = test, session = session, simul = simul, step = step, mono = mono}
fun updSanity  {input, output, lib, time, sub, sanity = _, nuprl, obid, ascii, tcheck, parse, split, prt, eval, alldef, test, session, simul, step, mono} sanity  = {input = input, output = output, lib = lib, time = time, sub = sub, sanity = sanity, nuprl = nuprl, obid = obid, ascii = ascii, tcheck = tcheck, parse = parse, split = split, prt = prt, eval = eval, alldef = alldef, test = test, session = session, simul = simul, step = step, mono = mono}
fun updNuprl   {input, output, lib, time, sub, sanity, nuprl = _, obid, ascii, tcheck, parse, split, prt, eval, alldef, test, session, simul, step, mono} nuprl   = {input = input, output = output, lib = lib, time = time, sub = sub, sanity = sanity, nuprl = nuprl, obid = obid, ascii = ascii, tcheck = tcheck, parse = parse, split = split, prt = prt, eval = eval, alldef = alldef, test = test, session = session, simul = simul, step = step, mono = mono}
fun updObid    {input, output, lib, time, sub, sanity, nuprl, obid = _, ascii, tcheck, parse, split, prt, eval, alldef, test, session, simul, step, mono} obid    = {input = input, output = output, lib = lib, time = time, sub = sub, sanity = sanity, nuprl = nuprl, obid = obid, ascii = ascii, tcheck = tcheck, parse = parse, split = split, prt = prt, eval = eval, alldef = alldef, test = test, session = session, simul = simul, step = step, mono = mono}
fun updAscii   {input, output, lib, time, sub, sanity, nuprl, obid, ascii = _, tcheck, parse, split, prt, eval, alldef, test, session, simul, step, mono} ascii   = {input = input, output = output, lib = lib, time = time, sub = sub, sanity = sanity, nuprl = nuprl, obid = obid, ascii = ascii, tcheck = tcheck, parse = parse, split = split, prt = prt, eval = eval, alldef = alldef, test = test, session = session, simul = simul, step = step, mono = mono}
fun updTcheck  {input, output, lib, time, sub, sanity, nuprl, obid, ascii, tcheck = _, parse, split, prt, eval, alldef, test, session, simul, step, mono} tcheck  = {input = input, output = output, lib = lib, time = time, sub = sub, sanity = sanity, nuprl = nuprl, obid = obid, ascii = ascii, tcheck = tcheck, parse = parse, split = split, prt = prt, eval = eval, alldef = alldef, test = test, session = session, simul = simul, step = step, mono = mono}
fun updParse   {input, output, lib, time, sub, sanity, nuprl, obid, ascii, tcheck, parse = _, split, prt, eval, alldef, test, session, simul, step, mono} parse   = {input = input, output = output, lib = lib, time = time, sub = sub, sanity = sanity, nuprl = nuprl, obid = obid, ascii = ascii, tcheck = tcheck, parse = parse, split = split, prt = prt, eval = eval, alldef = alldef, test = test, session = session, simul = simul, step = step, mono = mono}
fun updSplit   {input, output, lib, time, sub, sanity, nuprl, obid, ascii, tcheck, parse, split = _, prt, eval, alldef, test, session, simul, step, mono} split   = {input = input, output = output, lib = lib, time = time, sub = sub, sanity = sanity, nuprl = nuprl, obid = obid, ascii = ascii, tcheck = tcheck, parse = parse, split = split, prt = prt, eval = eval, alldef = alldef, test = test, session = session, simul = simul, step = step, mono = mono}
fun updPrint   {input, output, lib, time, sub, sanity, nuprl, obid, ascii, tcheck, parse, split, prt = _, eval, alldef, test, session, simul, step, mono} prt     = {input = input, output = output, lib = lib, time = time, sub = sub, sanity = sanity, nuprl = nuprl, obid = obid, ascii = ascii, tcheck = tcheck, parse = parse, split = split, prt = prt, eval = eval, alldef = alldef, test = test, session = session, simul = simul, step = step, mono = mono}
fun updEval    {input, output, lib, time, sub, sanity, nuprl, obid, ascii, tcheck, parse, split, prt, eval = _, alldef, test, session, simul, step, mono} eval    = {input = input, output = output, lib = lib, time = time, sub = sub, sanity = sanity, nuprl = nuprl, obid = obid, ascii = ascii, tcheck = tcheck, parse = parse, split = split, prt = prt, eval = eval, alldef = alldef, test = test, session = session, simul = simul, step = step, mono = mono}
fun updAlldef  {input, output, lib, time, sub, sanity, nuprl, obid, ascii, tcheck, parse, split, prt, eval, alldef = _, test, session, simul, step, mono} alldef  = {input = input, output = output, lib = lib, time = time, sub = sub, sanity = sanity, nuprl = nuprl, obid = obid, ascii = ascii, tcheck = tcheck, parse = parse, split = split, prt = prt, eval = eval, alldef = alldef, test = test, session = session, simul = simul, step = step, mono = mono}
fun updTest    {input, output, lib, time, sub, sanity, nuprl, obid, ascii, tcheck, parse, split, prt, eval, alldef, test = _, session, simul, step, mono} test    = {input = input, output = output, lib = lib, time = time, sub = sub, sanity = sanity, nuprl = nuprl, obid = obid, ascii = ascii, tcheck = tcheck, parse = parse, split = split, prt = prt, eval = eval, alldef = alldef, test = test, session = session, simul = simul, step = step, mono = mono}
fun updSession {input, output, lib, time, sub, sanity, nuprl, obid, ascii, tcheck, parse, split, prt, eval, alldef, test, session = _, simul, step, mono} session = {input = input, output = output, lib = lib, time = time, sub = sub, sanity = sanity, nuprl = nuprl, obid = obid, ascii = ascii, tcheck = tcheck, parse = parse, split = split, prt = prt, eval = eval, alldef = alldef, test = test, session = session, simul = simul, step = step, mono = mono}
fun updSimul   {input, output, lib, time, sub, sanity, nuprl, obid, ascii, tcheck, parse, split, prt, eval, alldef, test, session, simul = _, step, mono} simul   = {input = input, output = output, lib = lib, time = time, sub = sub, sanity = sanity, nuprl = nuprl, obid = obid, ascii = ascii, tcheck = tcheck, parse = parse, split = split, prt = prt, eval = eval, alldef = alldef, test = test, session = session, simul = simul, step = step, mono = mono}
fun updStep    {input, output, lib, time, sub, sanity, nuprl, obid, ascii, tcheck, parse, split, prt, eval, alldef, test, session, simul, step = _, mono} step    = {input = input, output = output, lib = lib, time = time, sub = sub, sanity = sanity, nuprl = nuprl, obid = obid, ascii = ascii, tcheck = tcheck, parse = parse, split = split, prt = prt, eval = eval, alldef = alldef, test = test, session = session, simul = simul, step = step, mono = mono}
fun updMono    {input, output, lib, time, sub, sanity, nuprl, obid, ascii, tcheck, parse, split, prt, eval, alldef, test, session, simul, step, mono = _} mono    = {input = input, output = output, lib = lib, time = time, sub = sub, sanity = sanity, nuprl = nuprl, obid = obid, ascii = ascii, tcheck = tcheck, parse = parse, split = split, prt = prt, eval = eval, alldef = alldef, test = test, session = session, simul = simul, step = step, mono = mono}

datatype arg = I       of string        (* input     *)
	     | O       of string        (* ouput     *)
	     | L       of string        (* library   *)
	     | T       of LargeInt.int  (* timelimit *)
	     | OBID    of string        (* object_id to use when generating the nuprl code *)
	     | EVAL    of string        (* term to evaluate *)
	     | DEF     of string        (* the alldef file  *)
	     | TEST    of int
	     | STEP    of int           (* to step through a protocl, by specifying (int) the message to send next *)
	     | SIMUL   of string        (* configuration file  *)
	     | SESSION
	     | TONUPRL
	     | SUBTYPING
	     | SANITY
	     | FROMASCII
	     | TYPECHECK
	     | PARSE
	     | SPLIT
	     | PRINT
	     | MONO

fun format args =
    let fun gen [] r = r
	  | gen (SUBTYPING      :: list) r = gen list (updSub     r true)
	  | gen (SANITY         :: list) r = gen list (updSanity  r true)
	  | gen (FROMASCII      :: list) r = gen list (updAscii   r true)
	  | gen (TYPECHECK      :: list) r = gen list (updTcheck  r true)
	  | gen (PARSE          :: list) r = gen list (updParse   r true)
	  | gen (SPLIT          :: list) r = gen list (updSplit   r true)
	  | gen (PRINT          :: list) r = gen list (updPrint   r true)
	  | gen (TONUPRL        :: list) r = gen list (updNuprl   r true)
	  | gen (SESSION        :: list) r = gen list (updSession r true)
	  | gen (MONO           :: list) r = gen list (updMono    r true)
	  | gen ((OBID    obid) :: list) r = gen list (updObid    r obid)
	  | gen ((T       time) :: list) r = gen list (updTime    r time)
	  | gen ((I       file) :: list) r = gen list (updInput   r file)
	  | gen ((SIMUL   file) :: list) r = gen list (updSimul   r (SOME file))
	  | gen ((STEP    n)    :: list) r = gen list (updStep    r (SOME n))
	  | gen ((TEST    n)    :: list) r = gen list (updTest    r (SOME n))
	  | gen ((EVAL    str)  :: list) r = gen list (updEval    r (SOME str))
	  | gen ((O       file) :: list) r = gen list (updOutput  r (SOME file))
	  | gen ((L       file) :: list) r = gen list (updLib     r (SOME file))
	  | gen ((DEF     file) :: list) r = gen list (updAlldef  r (SOME file))
    in gen args initArgs
    end


(* ------ A FEW GLOBAL VARIABLES ------ *)

val program = ref NONE : (string list * NT.nuprl_term) option ref

val default_alldefs = "/usr/fdl/lib/alldefs"

val loaded     = ref false
val configured = ref false

val components = ref [] : (NT.nuprl_term * NT.nuprl_term) list ref
val intransit  = ref [] : NT.nuprl_term list ref


(* ------ INTERFACE FUNCTIONS ------ *)

type timer = {real : Timer.real_timer,
	      cpu  : Timer.cpu_timer}

fun startTimer () = {real = Timer.startRealTimer (),
		     cpu  = Timer.startCPUTimer ()}
fun getTime (timer : timer) = Timer.checkRealTimer (#real timer)
fun getMilliTime timer = Time.toMilliseconds (getTime timer)

fun eval_wrap ev ts n t =
    let val timer = startTimer ()
	val res   = ev ts n t
	val time  = getMilliTime timer
	val _     = print ("[" ^ LargeInt.toString time ^ "ms]\n")
    in res
    end

val ev1  = EV.run_ev1_map
val ev2  = EV.run_ev2_map
val ev2b = EV.run_ev2b_map
val ev3  = EV.run_ev3_map
val ev3b = EV.run_ev3b_map
val ev4  = EV.run_ev4_map
val ev4b = EV.run_ev4b_map

val ev1  = eval_wrap ev1
val ev2  = eval_wrap ev2
val ev2b = eval_wrap ev2b
val ev3  = eval_wrap ev3
val ev3b = eval_wrap ev3b
val ev4  = eval_wrap ev4
val ev4b = eval_wrap ev4b

fun parseEML input =
    let val _    = D.reset ()
	val term = P.parse [input]
	val st1  = A.export term
	val st2  = A.toString term
	val _    = print (st1 ^ "\n")
	val _    = print (st2 ^ "\n")
    in ()
    end

fun getLastAbstraction [] = (NONE, [])
  | getLastAbstraction lst =
    let val rev    = List.rev lst
	val last   = List.hd rev
	val firsts = List.rev (List.tl rev)
    in if NT.is_nuprl_iabstraction_term last
       then (SOME (NT.dest_iabstraction last), [])
       else let val (termop, terms) = getLastAbstraction firsts
	    in (termop, terms @ [last])
	    end
    end

fun keepAbsWf [] = []
  | keepAbsWf (term :: terms) =
    if NT.is_nuprl_iabstraction_term term
       orelse
       NT.is_nuprl_iwftheorem_term term
    then term :: keepAbsWf terms
    else keepAbsWf terms

fun pairAbsWf [] = []
  | pairAbsWf [x] = raise EH.Impossible "pairAbsWf"
  | pairAbsWf (abs :: wf :: terms) =
    if NT.is_nuprl_iabstraction_term abs
       andalso
       NT.is_nuprl_iwftheorem_term wf
    then (abs, wf) :: pairAbsWf terms
    else raise EH.Impossible "pairAbsWf"

fun splitLastAbs terms =
    let val terms' = keepAbsWf terms
	val pairs  = pairAbsWf terms'
	val rev    = List.rev pairs
    in if List.length rev > 0
       then let val (abs, wf) = List.hd rev
		val firsts = List.rev (List.tl rev)
	    in (SOME (NT.dest_iabstraction abs, wf), firsts)
	    end
       else (NONE, [])
    end

fun evaluate_gen ev input str libop obid sub prt btyp =
    let val _        = D.reset ()
	val tmp      = OS.FileSys.tmpName ()
	val stout    = TextIO.openOut tmp
	val _        = TextIO.output (stout, "let it = " ^ str ^ "\n")
	val _        = TextIO.closeOut stout
	val san      = true
    in let val (term, env, b) = E.slice [input, tmp] NONE libop 1 prt sub san btyp
       in if b
	  then let val s_term        = A.simplify term
		   val (n_terms1, _) = N.toNuprl NONE tmp obid s_term env sub false true true btyp
		   val _             = OS.FileSys.remove tmp
	       in case splitLastAbs n_terms1 of
		      (SOME ((t1, t2, t3), wf), terms) =>
		      let val t3' = ev (NT.mk_nuprl_evalall_term t3) terms
			  val abs = NT.mk_nuprl_iabstraction_term t2 t3'
			  val rhs = #3 (NT.dest_iabstraction abs)
			  val thm = NT.dest_iwftheorem wf
			  val sta = NT.nuprl2eml_abs "it" abs
			  val stw = NT.nuprl2eml_wf  "it" wf
			  (*val _ = print (NT.toStringTerm wf)*)
			  val _   = if sta = "`???`"
				    then print ("evaluation failed.\n")
				    else print (sta ^ " : " ^ stw ^ "\n")
		      (*app (fn t => print (NT.nuprl2eml t ^ "\n")) [abs, wf]*)
		      in ()
		      end
		    | _ => ()
	       end
	  else print "Untypable.\n"
       end handle e => (OS.FileSys.remove tmp; raise e)
    end

fun evaluate input str libop alldefop obid sub =
    evaluate_gen (fn t => fn ts => #1 (EV.run_ev1 true alldefop ts (~1) t))
		 input str libop obid sub (true, true) true

fun evaluate_map ev input str libop obid sub btyp =
    evaluate_gen (fn t => fn ts => let val (u, n) = ev ts (~1) t in u end)
		 input str libop obid sub (true, false) btyp

fun run_gen_session ev str strm libop input obid sub btyp =
    let val _ = print ("\nswitched to evaluator" ^ str ^ "\n")
	val _ = print "EventML# "
    in run_session ev strm libop input obid sub btyp
    end

and run_session ev strm libop input obid sub btyp =
    case TextIO.inputLine strm of
	NONE => run_session ev strm libop input obid sub btyp
      | SOME "quit\n"   => (print "\nend of session\n"; EV.end_session ())
      | SOME "exit\n"   => (print "\nend of session\n"; EV.end_session ())
      | SOME "type\n"   => (print "\nenabled type inferencer\n";  print "EventML# "; run_session ev strm libop input obid sub true)
      | SOME "notype\n" => (print "\ndisabled type inferencer\n"; print "EventML# "; run_session ev strm libop input obid sub false)
      | SOME "ev1\n"    => run_gen_session ev1  "1"  strm libop input obid sub btyp
      | SOME "ev2\n"    => run_gen_session ev2  "2"  strm libop input obid sub btyp
      | SOME "ev2b\n"   => run_gen_session ev2b "2b" strm libop input obid sub btyp
      | SOME "2b\n"     => run_gen_session ev2b "2b" strm libop input obid sub btyp
      | SOME "ev3\n"    => run_gen_session ev3  "3"  strm libop input obid sub btyp
      | SOME "ev3b\n"   => run_gen_session ev3b "3b" strm libop input obid sub btyp
      | SOME "3b\n"     => run_gen_session ev3b "3b" strm libop input obid sub btyp
      | SOME "ev4\n"    => run_gen_session ev4  "4"  strm libop input obid sub btyp
      | SOME "ev4b\n"   => run_gen_session ev4b "4b" strm libop input obid sub btyp
      | SOME "4b\n"     => run_gen_session ev4b "4b" strm libop input obid sub btyp
      | SOME line       =>
	let val _ = evaluate_map ev input line libop obid sub btyp
	    val _ = print "EventML# "
	in run_session ev strm libop input obid sub btyp
	end

fun start_session input libop alldefop obid sub =
    let val _    = print "Loading library...\n"
	val _    = EV.start_session false alldefop
	val strm = TextIO.stdIn
	val _    = print "EventML# "
	val btyp = true
    in run_session ev1 strm libop input obid sub btyp
    end

val eval_test = EV.test

val typecheck = E.typecheck

fun slice input outputop libop timelimit sub =
    let val san  = false
	val btyp = true
	val (prog, env, b) = E.slice [input] outputop libop timelimit (true, true) sub san btyp
    in ()
    end

fun toNuprl input outputop libop alldefop obid timelimit sub prt poly btyp =
    let val san = false
	val (term, env, b) = E.slice [input] NONE libop timelimit (true, true) sub san btyp
    in if b
       then let val s_term = A.simplify term
		val mapop =
		    case alldefop of
			SOME f =>
			let val lib = NT.terms2map (PN.parse prt f false)
			    val _   = EV.start_session_lib lib
			    val _   = loaded := true
			in SOME lib
			end
		      | NONE => NONE
		val (n_terms, prog) = N.toNuprl mapop input obid s_term env sub true poly true btyp
		val _  = program := prog
		val st =
		    if prt
		    then app (fn n_term => print (NT.toStringTerm n_term ^ "\n")) n_terms
		    else ()
		val _  =
		    case outputop of
			NONE => ()
		      | SOME file =>
			let val file' = file ^ ".tmp"
			    val stout = TextIO.openOut file'
			    val _     = map (fn n_term => TextIO.output (stout, NT.toStringTerm n_term ^ "\004\n")) n_terms
			    val _     = TextIO.closeOut stout
			    val _     = OS.FileSys.rename {old = file', new = file}
			in () end
			handle IO.Io {name, function, cause} =>
			       (raise EH.Impossible ("cannot open or close one of the output file (function:" ^ function ^ ",name:" ^ name ^ ")\n"))
	    in n_terms
	    end
       else (print "We won't convert the program into a Nuprl term because it contains errors.\n"; [])
    end

fun parseNuprl input outputop prt split =
    GL.exportTerms input outputop prt split

fun sanitizer input outputop libop timelimit sub =
    let val san  = true
	val btyp = true
	val (prog, env, b) = E.slice [input] outputop libop timelimit (true, true) sub san btyp
    in ()
    end

fun getComponents term =
    if NT.is_nuprl_cons_term term
    then let val (head, tail) = NT.dest_cons term
	 in if NT.is_nuprl_pair_term head
	    then (NT.dest_pair 1 head) :: (getComponents tail)
	    else raise EH.Impossible "getComponents"
	 end
    else if NT.is_nuprl_nil_term term
    then []
    else raise EH.Impossible "getComponents"

fun how_many_components term =
    if NT.is_nuprl_pair_term term
    then 1 + how_many_components (#2 (NT.dest_pair 2 term))
    else 1

fun all t   =
    if NT.is_nuprl_iclosure_term t
    then let val (a,e) = NT.dest_iclosure t
	 in NT.mk_nuprl_iclosure_term (all a) e
	 end
    else NT.mk_nuprl_evalall_term t
fun app p t =
    if NT.is_nuprl_iclosure_term p
    then let val (a,e) = NT.dest_iclosure p
	 in NT.mk_nuprl_iclosure_term (app a t) e
	 end
    else NT.mk_nuprl_apply_term p t

fun testing 0 = EV.start_session true (SOME default_alldefs)
  | testing 1 = EV.end_session ()
  | testing n =
    let val terms   = GL.parseTerms "output" false false
	(* list of locations: *)
	val pid     = NT.mk_nuprl_mkid_term "p"
	val l1id    = NT.mk_nuprl_mkid_term "l1"
	val l2id    = NT.mk_nuprl_mkid_term "l2"
	val ids     = [pid, l1id, l2id]
	val lstls   = NT.mk_nuprl_finite_list_term [l1id, l2id]
    in case terms of
	   [prog] =>
	   let (* program: *)
	       val prog1 = NT.mk_nuprl_apply_term prog lstls
	       val prog2 = NT.mk_nuprl_apply_term prog1 pid
	       val prog3 = #1 (ev1 [] (~1) prog2)
	       fun build id = NT.mk_nuprl_df_program_meaning_term (app prog3 id)
	       (*val xxx =
		   let val x = ev n (build pid)
		       val _ = print ("\n-+1+-----------\n" ^ NT.toStringTerm prog  ^ "\n-+1+-----------\n")
		       val _ = print ("\n-+2+-----------\n" ^ NT.toStringTerm prog3 ^ "\n-+2+-----------\n")
		       val _ = print ("\n-+3+-----------\n" ^ NT.toStringTerm x     ^ "\n-+3+-----------\n")
		       val _ = raise Fail ""
		   in ()
		   end*)
	       val lst = map (fn id => (id, #1 (ev1 [] (~1) (build id)))) ids
	       (* inital message: *)
	       val client = NT.mk_nuprl_mkid_term "client"
	       val start  = NT.mk_nuprl_token_term "start"
	       val hdr    = NT.mk_nuprl_finite_list_term [start]
	       val loc    = NT.mk_nuprl_loc_term
	       val msg    = NT.mk_nuprl_make_msg_term hdr loc client
	   (* now we select the program associated with "p" *)
	   in case List.find (fn (id, p) => NT.alpha_equal_terms id pid) lst of
		  SOME (i, p) =>
		  let (*val _ = print ("\n-+a+-----------\n" ^ NT.toStringTerm p  ^ "\n-+a+-----------\n")*)
		      val p' = #1 (ev1 [] (~1) (all (app p msg)))
		  in print ("\n-+b+-----------\n" ^ NT.toStringTerm p' ^ "\n-+b+-----------\n")
		  end
		| NONE => ()
	   end
	 | _ => ()
    end

fun extractInfoParam list =
    let val locs =
	    map (fn loc =>
		    if String.isPrefix "LOC(" loc
		       andalso
		       String.isSuffix ")" loc
		    then let val to = String.size loc - 5
			     val id = String.substring (loc, 4, to)
			 in NT.mk_nuprl_mkid_term id
			 end
		    else raise Fail ("extract_info_param - " ^ loc))
		list
    in case locs of
	   []  => raise Fail ("extract_info_param - " ^ Int.toString (List.length locs))
	 | [l] => l
	 | _   => NT.mk_nuprl_finite_list_term locs
    end

(*
(* Write a proper parser for that! *)
fun load_config config =
    let val strm = TextIO.openIn config
	val fsep = fn #" " => true | _ => false
	val locations =
	    case TextIO.inputLine strm of
		SOME line =>
		let val line = String.substring (line, 0, String.size line - 1)
		    val elts = String.tokens fsep line
		in case elts of
		       x :: xs =>
		       if x = "locations:"
		       then map NT.mk_nuprl_mkid_term xs
		       else raise Fail "load_config"
		     | [] => raise Fail "load_config"
		end
	      | NONE => raise Fail "load_config"
	fun getParams () =
	    case TextIO.inputLine strm of
		SOME line =>
		let val line = String.substring (line, 0, String.size line - 1)
		    val elts = String.fields fsep line
		in case elts of
		       (name :: vs) =>
		       if name = ""
		       then getParams ()
		       else (case String.fields (fn #":" => true | _ => false) name of
				 [name, ""] => (name, extractInfoParam vs) :: getParams ()
			       | _ => raise Fail "load_config - name")
		     | _ => raise Fail "load_config - elts"
		end
	      | NONE => []
	val params = getParams ()
	val _ = TextIO.closeIn strm
    in (locations, params, [])
    end
*)

fun load_config_parse config =
    let val (ls, ps, ms) = C.parse config
	val locations = map (fn (loc, host, port) => NT.mk_nuprl_mkid_term loc) ls
	val messages  =
	    map (fn (id, (hdr, typ, body)) =>
		    let val msg = NT.mk_nuprl_make_msg_term hdr typ body
			val pid = NT.mk_nuprl_mkid_term id
		    in NT.mk_nuprl_pair_term pid msg
		    end)
		ms
    in (locations, ps, messages)
    end

fun testing _ = (load_config_parse "new_conf"; ())

fun create_listening_socket () =
    let val sock      = INetSock.TCP.socket ()
	val addr      = Option.valOf (NetHostDB.fromString ("127.0.0.1"))
	val port      = 8987
	val sock_addr = INetSock.toAddr (addr, port)
	val _         = Socket.bind (sock, sock_addr)
	val _         = Socket.listen (sock, 100)
    in (sock, sock_addr)
    end

fun testing _ =
    let val sock      = INetSock.TCP.socket ()
	(*----*)
	val addr      = Option.valOf (NetHostDB.fromString ("127.0.0.1"))
	val port      = 8989
	val sock_addr = INetSock.toAddr (addr, port)
	(*----*)
	val (lsn,adr) = create_listening_socket ()
	(*----*)
	val vect      = Word8Vector.fromList [0w1]
	val slice     = Word8VectorSlice.full vect
	val _         = print (".\n")
	val _         = Socket.connect (sock, adr)
	val _         = print ("..\n")
	val _         = Socket.sendVec (sock, slice)
	val _         = print ("...\n")
    in ()
    end handle _ => raise Fail "error"

fun testing n =
    let val _ = EV.start_session true (SOME default_alldefs)
	val _ = print ("[loaded library]")
    in eval_test n
    end

fun getParams params map_params =
    map (fn param =>
	    case List.find (fn (id : string, _) => id = param) map_params of
		SOME (id, v) => v
	      | NONE => raise Fail ("get_params(" ^ param ^ ")"))
	params

fun update_component (i, s) cmps =
    map (fn (j, p) =>
	    ((*print ("--" ^ NT.toStringTerm j ^ "\n");*)
	     if NT.alpha_equal_terms i j
	     then (i, s)
	     else (j, p)))
	cmps

fun print_messages_intransit () =
    let fun aux n [] = ()
	  | aux n (t :: ts) =
	    let val _ = print (Int.toString n ^ ": " ^ NT.nuprlTerm2eml t ^ "\n--\n")
	    in aux (n + 1) ts
	    end
    in print ("\n-------IN-TRANSIT-------\n"); aux 1 (!intransit)
    end

fun debug_eval previous_result term n =
    let val _ = print ("-" ^ Int.toString n ^ "\n")
	val (p1,n1) = ev1 [] n term
	val (p2,n2) = ev2 [] n term
    in if n1 = n2 andalso NT.alpha_equal_terms p1 p2
       then if n1 = n
	    then debug_eval p1 term (n + 1)
	    else print ("equal[1:" ^ Int.toString n1 ^ ",2:" ^ Int.toString n2 ^ "]\n")
       else let val stout = TextIO.openOut "output-old"
		val _     = TextIO.output (stout, NT.toStringTerm previous_result ^ "\004\n")
		val _     = TextIO.closeOut stout
		val stout = TextIO.openOut "output-1"
		val _     = TextIO.output (stout, NT.toStringTerm p1 ^ "\004\n")
		val _     = TextIO.closeOut stout
		val stout = TextIO.openOut "output-2"
		val _     = TextIO.output (stout, NT.toStringTerm p2 ^ "\004\n")
		val _     = TextIO.closeOut stout
	    in print ("diff[1:" ^ Int.toString n1 ^ ",2:" ^ Int.toString n2 ^ "]\n")
	    end
    end

fun run_stepper n ev debug config alldefs =
    if n < 0
    then let val _ = configured := false
	     val _ = components := []
	     val _ = intransit  := []
	 in ()
	 end
    else if !loaded
    then if !configured
	 then case !intransit of
		  [] => print "[no message in transit]\n"
		| lst =>
		  if n > List.length lst orelse n < 1
		  then print "[out of bound]\n"
		  else let val (id, msg) = NT.dest_pair 3 (List.nth (lst, n - 1))
			   (*val _ = print ("-+-" ^ NT.toStringTerm id ^ "\n")*)
		       in case List.find (fn (i, p) =>
					     ((*print ("---" ^ NT.toStringTerm i ^ "\n");*)
					      NT.alpha_equal_terms i id))
					 (!components) of
			      SOME (i, p) =>
			      let (*val _ = print ("\n-+a+-----------\n" ^ NT.toStringTerm p  ^ "\n-+a+-----------\n")*)
				  val toeval = all (app p msg)
				  val (p',stps) = ev [] (~1) toeval
				  val _ = if debug
					  then debug_eval toeval toeval 2025
					  else () (**)
				  (*val _ = raise Fail "--"*)
				  val (s,msgs) = NT.dest_pair 4 p'
				  (*val _ = print ("\n-+b+-----------\n" ^ NT.toStringTerm p' ^ "\n-+b+-----------\n")*)
				  (*val _ = print ("\n-------MSGS------\n" ^ NT.toStringTerm msgs ^ "\n------------\n")*)
				  val lst_msgs = NT.dest_list msgs
				  val new_cmps = update_component (i, s) (!components)
				  val new_msgs = (List.take (lst, n - 1)) @ (List.drop (lst, n)) @ lst_msgs
				  val _ = components := new_cmps
				  val _ = intransit := new_msgs
			      in ()
			      end
			    | NONE => print "[no recipient]\n"
		       end
	 else let (* gets the locations and parameters from the input config file *)
		 val (locations, params, messages) = load_config_parse config
		 (*val (locations, params, messages) = load_config config*)
		 (* generates the list of components *)
		 val cmps =
		     case !program of
			 SOME (prms, p) =>
			 let val args  = getParams prms params
			     val prog1 = NT.mk_nuprl_applies_term p args
			     val (prog2,stps) = ev [] (~1) prog1
			     fun build id = NT.mk_nuprl_df_program_meaning_term (app prog2 id)
			 in map (fn id => (id, let val (p,stps) = ev [] (~1) (build id)
						   (*val _ = print (NT.toStringTerm p)*)
					       in p
					       end))
				locations
			 end
		       | NONE => raise Fail "[not a program]"
		 val _ = components := cmps
		 val _ = intransit  := messages
		 val _ = configured := true
		 val _ = print ("[components loaded]\n")
	     in ()
	     end
    else let (* loads the alldefs file *)
	    val _ = EV.start_session true (SOME default_alldefs)
	    val _ = loaded := true
	    val _ = print ("[library loaded]\n")
	in ()
	end

fun loop_gen_stepper ev str strm conf alldefs =
    let val _ = print ("\nswitched to evaluator" ^ str ^ "\n")
	val _ = print "message? "
    in loop_stepper ev strm conf alldefs
    end

and loop_stepper ev strm conf alldefs =
    case TextIO.inputLine strm of
	NONE => loop_stepper ev strm conf alldefs
      | SOME "quit\n" =>
	let val _ = print "\nend of session\n"
	in run_stepper (~1) ev false conf alldefs
	end
      | SOME "exit\n" =>
	let val _ = print "\nend of session\n"
	in run_stepper (~1) ev false conf alldefs
	end
      | SOME "ev1\n"  => loop_gen_stepper ev1  "1"  strm conf alldefs
      | SOME "ev2\n"  => loop_gen_stepper ev2  "2"  strm conf alldefs
      | SOME "ev2b\n" => loop_gen_stepper ev2b "2b" strm conf alldefs
      | SOME "2b\n"   => loop_gen_stepper ev2b "2b" strm conf alldefs
      | SOME "ev3\n"  => loop_gen_stepper ev3  "3"  strm conf alldefs
      | SOME "ev3b\n" => loop_gen_stepper ev3b "3b" strm conf alldefs
      | SOME "3b\n"   => loop_gen_stepper ev3b "3b" strm conf alldefs
      | SOME "ev4\n"  => loop_gen_stepper ev4  "4"  strm conf alldefs
      | SOME "ev4b\n" => loop_gen_stepper ev4b "4b" strm conf alldefs
      | SOME "4b\n"   => loop_gen_stepper ev4b "4b" strm conf alldefs
      | SOME line =>
	let fun aux [] =
		(print_messages_intransit ();
		 print "message? ";
		 loop_stepper ev strm conf alldefs)
	      | aux (input :: inputs) =
		let val (input',debug) =
			if String.isPrefix "d" input
			then (String.substring (input, 1, (String.size input) - 1), true)
			else (input, false)
		in case Int.fromString input' of
		       SOME n => (run_stepper n ev debug conf alldefs;
				  aux inputs)
		     | NONE => (print "please, enter a nat\n";
				print "message? ";
				loop_stepper ev strm conf alldefs)
		end
	in aux (String.tokens (fn #" " => true | #"\n" => true | _ => false) line)
	end (*handle _ => print "empty line"*)

fun start_stepper spec conf lib alldefs obid time sub prt =
    let (* loads library and generates program *)
	val _     = print ("[type checking and loading library]\n")
	val poly  = true
	val btyp  = true
	val terms = toNuprl spec NONE lib alldefs obid time sub prt poly btyp
	val _     = EV.add_to_session terms
	(* generates stream on the standard input *)
	val strm  = TextIO.stdIn
	(* choose evaluator *)
	val _     = print "evaluator(1/2/2b/3/3b/4/4b)? "
	val ev    =
	    case TextIO.inputLine strm of
		NONE          => (print "evalutor1 has been selected\n";  ev1)
	      | SOME "ev2\n"  => (print "evalutor2 has been selected\n";  ev2)
	      | SOME "2\n"    => (print "evalutor2 has been selected\n";  ev2)
	      | SOME "ev2b\n" => (print "evalutor2b has been selected\n"; ev2b)
	      | SOME "2b\n"   => (print "evalutor2b has been selected\n"; ev2b)
	      | SOME "ev3\n"  => (print "evalutor3 has been selected\n";  ev3)
	      | SOME "3\n"    => (print "evalutor3 has been selected\n";  ev3)
	      | SOME "ev3b\n" => (print "evalutor3b has been selected\n"; ev3b)
	      | SOME "3b\n"   => (print "evalutor3b has been selected\n"; ev3b)
	      | SOME "ev4\n"  => (print "evalutor4 has been selected\n";  ev4)
	      | SOME "4\n"    => (print "evalutor4 has been selected\n";  ev4)
	      | SOME "ev4b\n" => (print "evalutor4b has been selected\n"; ev4b)
	      | SOME "4b\n"   => (print "evalutor4b has been selected\n"; ev4b)
	      | SOME line     => (print "evalutor1 has been selected\n"; ev1)
	(* loads the components *)
	val _     = print ("[loading components]\n")
	val _     = run_stepper 0 ev false conf alldefs
	(* print the list of messages initially in transit *)
	val _     = print ("[running simulator]\n")
	val _     = print_messages_intransit ()
	val _     = print "message? "
    (* starts simulator *)
    in loop_stepper ev strm conf alldefs
    end

fun run args =
    let val {input, output, lib, time, sub, sanity, nuprl, obid, ascii, tcheck, parse, split, prt, eval, alldef, test, session, simul, step, mono} = format args
	val poly = not mono
	val btyp = true
    in if Option.isSome test
       then testing (Option.valOf test)
       else if parse
       then parseEML input
       else if tcheck
       then typecheck input sub
       else if sanity
       then sanitizer input output lib time sub
       else if nuprl
       then (toNuprl input output lib alldef obid time sub prt poly btyp; ())
       else if ascii
       then parseNuprl input output prt split
       else if Option.isSome eval
       then evaluate input (Option.valOf eval) lib alldef obid sub
       else if session
       then start_session input lib alldef obid sub
       else if Option.isSome step
       then run_stepper (Option.valOf step) ev1 false input alldef
       else if Option.isSome simul
       then start_stepper input (Option.valOf simul) lib alldef obid time sub prt
       else slice input output lib time sub
    end

end
